"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CacheItemState = void 0;
var CacheItemState;
(function (CacheItemState) {
    CacheItemState[CacheItemState["RESERVED"] = 0] = "RESERVED";
    CacheItemState[CacheItemState["PRESENT"] = 1] = "PRESENT";
})(CacheItemState = exports.CacheItemState || (exports.CacheItemState = {}));
class Cache {
    constructor({ itemLifetime, monitoringFrequency, waiterTimeout, waiterFrequency }) {
        this.cache = new Map();
        this.cacheLifetime = itemLifetime || 1800000;
        this.monitoringFrequency = monitoringFrequency || 60000;
        this.waiterTimeout = waiterTimeout || 30000;
        this.waiterFrequency = waiterFrequency || 100;
        this.isMonitoring = false;
    }
    runMonitor() {
        if (this.isMonitoring) {
            return;
        }
        this.isMonitoring = true;
        setTimeout(() => {
            const newCache = new Map();
            let cacheRenewed = false;
            Object.keys(this.cache).forEach((id) => {
                const item = this.cache.get(id);
                if (!this.isExpired(item)) {
                    newCache.set(id, item);
                    cacheRenewed = true;
                }
            });
            this.cache = newCache;
            this.isMonitoring = false;
            if (cacheRenewed) {
                this.runMonitor();
            }
        }, this.monitoringFrequency);
    }
    isExpired(cacheItem) {
        const now = new Date().getTime();
        const elapsed = now - cacheItem.createdAt;
        return elapsed > this.cacheLifetime;
    }
    getItem(id) {
        if (this.cache.get(id) === undefined) {
            return null;
        }
        const cached = this.cache.get(id);
        if (this.isExpired(cached)) {
            this.cache.delete(id);
            return;
        }
        return cached;
    }
    reserveWaiter(id, fetcher) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                const started = new Date().getTime();
                const resolveItem = () => __awaiter(this, void 0, void 0, function* () {
                    const item = this.cache.get(id);
                    if (typeof item === 'undefined') {
                        return yield this.reserveAndFetch(id, fetcher);
                    }
                    switch (item.state) {
                        case CacheItemState.RESERVED:
                            const elapsed = new Date().getTime() - started;
                            if (elapsed < this.waiterTimeout) {
                                waiter();
                            }
                            else {
                                reject(new Error(`Timed-out waiting for cache item to become available; timeout ${this.waiterTimeout}, waited ${elapsed}`));
                            }
                            break;
                        case CacheItemState.PRESENT:
                            resolve(item);
                    }
                });
                const waiter = () => __awaiter(this, void 0, void 0, function* () {
                    setTimeout(resolveItem, this.waiterFrequency);
                });
                waiter();
            });
        });
    }
    reserveAndFetch(id, fetcher) {
        return __awaiter(this, void 0, void 0, function* () {
            this.reserveItem(id, fetcher);
            const newItem = yield fetcher();
            const newCacheItem = {
                id, fetcher,
                createdAt: new Date().getTime(),
                value: newItem,
                state: CacheItemState.PRESENT
            };
            this.cache.set(id, newCacheItem);
            this.runMonitor();
            return newItem;
        });
    }
    getItemWithWait({ id, fetcher }) {
        return __awaiter(this, void 0, void 0, function* () {
            const cached = this.cache.get(id);
            if (typeof cached === 'undefined') {
                return this.reserveAndFetch(id, fetcher);
            }
            if (this.isExpired(cached)) {
                this.cache.delete(id);
                return this.reserveAndFetch(id, fetcher);
            }
            switch (cached.state) {
                case CacheItemState.RESERVED:
                    return (yield this.reserveWaiter(id, fetcher)).value;
                case CacheItemState.PRESENT:
                    return cached.value;
            }
        });
    }
    reserveItem(id, fetcher) {
        const reservedItem = {
            id, fetcher,
            reservedAt: new Date().getTime(),
            state: CacheItemState.RESERVED
        };
        this.cache.set(id, reservedItem);
        return reservedItem;
    }
}
exports.default = Cache;
//# sourceMappingURL=Cache.js.map